; --------------------------------------------------- ;
; SND.ASM                                             ;
;                                                     ;
; Written by: Geoff Friesen (November, 1991)          ;
;                                                     ;
; This TSR extends the BIOS to include sound support. ;
; Sound support is accessed via interrupt 1ah.        ;
;                                                     ;
; Syntax: SND [/U]                                    ;
;                                                     ;
; The /U option is used to uninstall SND.             ;
;                                                     ;
; This is version 1.  Four commands are implemented.  ;
; Register AH is loaded with FFh and register AL with ;
; the command code (0-3).  If command is not in range ;
; 0-3 then an error code of 255 is returned in AH.    ;
;                                                     ;
; Command 0: Get version number                       ;
;                                                     ;
; mov ax, ff00h                                       ;
; int 1ah                                             ;
;                                                     ;
; Exit: AL = 1 (version number)                       ;
;       AH = 0 (success)                              ;
;       AH = 1 (another command in progress)          ;
;                                                     ;
; Command 1: Generate a delay in timer ticks          ;
;                                                     ;
; mov ax, ff01h                                       ;
; mov dx, ticks                                       ;
; int 1ah                                             ;
;                                                     ;
; Exit: AH = 0 (success)                              ;
;       AH = 1 (another command in progress)          ;
;                                                     ;
; Command 2: Enable sound                             ;
;                                                     ;
; mov ax, ff02h                                       ;
; mov dx, frequency                                   ;
; int 1ah                                             ;
;                                                     ;
; Exit: AH = 0 (success)                              ;
;       AH = 1 (another command in progress)          ;
;       AH = 2 (frequency less than 20 HZ)            ;
;                                                     ;
; Command 3: Disable sound                            ;
;                                                     ;
; mov ax, ff03h                                       ;
; int 1ah                                             ;
;                                                     ;
; Exit: AH = 0 (success)                              ;
;       AH = 1 (another command in progress)          ;
; --------------------------------------------------- ;

CR              EQU     13
LF              EQU     10
TAB             EQU     9

FALSE           EQU     0
TRUE            EQU     1

MAXCMD          EQU     3
VER             EQU     1

_TEXT           SEGMENT BYTE PUBLIC 'CODE'
		ASSUME  cs:_TEXT, ds:_TEXT, es:_TEXT, ss:_TEXT
		ORG     100h
snd:
		jmp     NEAR PTR install

; ------------- ;
; Resident Data ;
; ------------- ;

; ------------------------------------------------------------------- ;
; Do not modify the signature "SND" or place any instructions or data ;
; between the "jmp NEAR PTR install" instruction and "signature".     ;
; ------------------------------------------------------------------- ;

signature       DB      "SND"           ; Used to check for prior install
LSIGNATURE      EQU     $-signature

active          DB      FALSE           ; Int 1ah active flag
oldint1ah       DD      ?               ; Old int 1ah vector

; ------------- ;
; Resident Code ;
; ------------- ;

newint1ah       PROC    FAR
		cmp     ah, 0ffh        ; SND code?
		jz      newint1ah1      ; Yes, branch.

		jmp     DWORD PTR cs:oldint1ah ; Exit to original handler.
newint1ah1:
		cmp     al, MAXCMD      ; Command in range?
		jbe     newint1ah2      ; Yes, branch.

		mov     ah, 255         ; Out of range error code.
		iret
newint1ah2:
		cmp     BYTE PTR cs:active, FALSE ; Are we active?
		jz      newint1ah3      ; No, branch.

		mov     ah, 1           ; Active error code.
		iret
newint1ah3:
		mov     BYTE PTR cs:active, TRUE ; Flag active.

		sti                     ; Enable interrupts.

		cmp     al, 0           ; Get version command?
		jnz     newint1ah4      ; No, branch.

		mov     al, VER         ; Version number.
		mov     ah, 0           ; OK code.
		jmp     SHORT exit
newint1ah4:
		cmp     al, 1           ; Delay command?
		jnz     newint1ah5      ; No, branch.

		call    delay           ; Delay for DX ticks.

		mov     ah, 0           ; OK code.
		jmp     SHORT exit
newint1ah5:
		cmp     al, 2           ; Enable sound command?
		jnz     newint1ah7      ; No, branch.

		cmp     dx, 20          ; Frequency < 20 HZ?
		jae     newint1ah6      ; No, branch.

		mov     ah, 2           ; Bad frequency error code.
		jmp     SHORT exit
newint1ah6:
		call    soundon         ; Enable sound.

		mov     ah, 0           ; OK code.
		jmp     SHORT exit
newint1ah7:
		call    soundoff        ; Disable sound.

		mov     ah, 0           ; OK code.
exit:
		cli
		mov     BYTE PTR cs:active, FALSE ; Flag inactive.
		iret
newint1ah       ENDP

; ------------------------------------------------------------ ;
; Local Procedure: delay                                       ;
;                                                              ;
; Delay for a specific number of timer ticks.  The BIOS memory ;
; locations 0000:046C and 0000:046D contain the low word of    ;
; the BIOS timer tick counter.  Timer ticks occur at a rate of ;
; approximately 18.2 ticks per second (unless this rate has    ;
; been reprogrammed).                                          ;
;                                                              ;
; Entry: DX - number of timer ticks to delay                   ;
;                                                              ;
; All registers preserved.                                     ;
; ------------------------------------------------------------ ;

delay           PROC    NEAR
		push    ax
		push    dx
		push    es

		xor     ax, ax
		mov     es, ax

		add     dx, es:[46ch]   ; Compute ending delay value.
delay1:
		cmp     dx, es:[46ch]   ; Compare current & ending values.
		jnz     delay1          ; Branch if current not = ending.

		pop     es
		pop     dx
		pop     ax
		ret
delay           ENDP

; ------------------------------------- ;
; Local Procedure: soundoff             ;
;                                       ;
; Disable sound by turning of the timer ;
; and speaker bits via port 61h.        ;
;                                       ;
; All registers preserved.              ;
; ------------------------------------- ;

soundoff        PROC    NEAR
		push    ax

		in      al, 61h         ; Turn off speaker/timer bits.
		and     al, 0fch
		out     61h, al

		pop     ax
		ret
soundoff        ENDP

; ------------------------------ ;
; Local Procedure: soundon       ;
;                                ;
; Entry: DX = frequency of sound ;
;                                ;
; Exit: sound is active          ;
;                                ;
; All registers preserved.       ;
; ------------------------------ ;

soundon         PROC    NEAR
		push    ax
		push    bx
		push    dx

		mov     bx, dx          ; Get frequency.
		mov     ax, 13533       ; Get 1193181 (input clock freq).
		mov     dx, 18
		div     bx              ; Calc channel C counter value.
					; Divide 1193181 by frequency.
		mov     bx, ax          ; Save quotient.

		in      al, 61h         ; Disable timer and speaker.
		and     al, 11111100B
		out     61h, al

		mov     al, 0b6h        ; Program command register.
		out     43h, al

		mov     al, bl          ; Load LSB of frequency value.
		out     42h, al

		mov     al, bh          ; Load MSB of frequency value.
		out     42h, al

		in      al, 61h         ; Enable timer and speaker.
		or      al, 00000011B
		out     61h, al

		pop     dx
		pop     bx
		pop     ax
		ret
soundon         ENDP

res_end         LABEL   BYTE

; ----------------- ;
; Installation Data ;
; ----------------- ;

errmsg1         DB      CR, LF
		DB      "snd: already installed"
		DB      CR, LF
		DB      '$'

errmsg2         DB      CR, LF
		DB      "snd: bad command line"
		DB      CR, LF
		DB      '$'

errmsg3         DB      CR, LF
		DB      "snd: not in memory"
		DB      CR, LF
		DB      '$'

errmsg4         DB      CR, LF
		DB      "snd: cannot uninstall"
		DB      CR, LF
		DB      '$'

okmsg1          DB      CR, LF
		DB      "snd: successfully installed"
		DB      CR, LF
		DB      '$'

okmsg2          DB      CR, LF
		DB      "snd: successfully uninstalled"
		DB      CR, LF
		DB      '$'

psp             DW      ?

snd_title       DB      CR, LF
		DB      "ͻ"
		DB      CR, LF
		DB      " SND v1: BIOS Sound Extension "
		DB      CR, LF
		DB      "                              "
		DB      CR, LF
		DB      " Public Domain Software       "
		DB      CR, LF
		DB      "                              "
		DB      CR, LF
		DB      " Written by: Geoff Friesen    "
		DB      CR, LF
		DB      "ͼ"
		DB      CR, LF
		DB      '$'

; ----------------- ;
; Installation Code ;
; ----------------- ;

install:
		sti
		cld

		mov     ah, 9           ; Display opening title.
		lea     dx, snd_title
		int     21h

		lea     bx, signature   ; Search for resident copy.
		not     BYTE PTR [bx]   ; Avoid false cache match.
		mov     cx, LSIGNATURE  ; Get length of signature.
		call    search          ; Search for prior copy.
		mov     psp, ax         ; Save PSP of prior/current TSR.

		mov     si, 81h         ; Point to command tail start.
install1:
		lodsb                   ; Get command tail character.

		cmp     al, ' '         ; Skip spaces.
		jz      install1

		cmp     al, TAB         ; Skip tabs.
		jz      install1

		cmp     al, CR          ; Exit on carriage return.
		jz      install5

		cmp     al, '/'         ; Start of "/U" parameter?
		jnz     install3        ; No, branch (error).

		lodsb                   ; Get parameter character.

		and     al, 0dfh        ; Make sure it is uppercase.

		cmp     al, 'U'         ; U option?
		jnz     install3        ; No, branch (error).

		lea     dx, errmsg3     ; Not in memory message.

		mov     ax, cs          ; CS=PSP means not in memory.
		cmp     ax, psp
		jz      install4        ; Branch if CS=PSP.

		lea     dx, errmsg4     ; Unsuccessful uninstall message.

		call    uninstall       ; Attempt to uninstall.
		jc      install4        ; Branch if unsuccessful.

		mov     ah, 9           ; Display ok uninstall message.
		lea     dx, okmsg2
		int     21h

		mov     ax, 4c00h       ; Exit with OK status code.
		int     21h
install3:
		lea     dx, errmsg2     ; Bad command line message.
install4:
		mov     ah, 9           ; Display message.
		int     21h

		mov     ax, 4c01h       ; Exit with ERROR status code.
		int     21h
install5:
		lea     dx, errmsg1     ; Already installed message

		mov     ax, cs          ; CS=PSP means not in memory.
		cmp     ax, psp
		jnz     install4        ; Branch if CS not = PSP.

		mov     ax, 351ah       ; Save current int 1ah vector.
		int     21h

		mov     WORD PTR oldint1ah, bx
		mov     WORD PTR oldint1ah [2], es

		mov     ax, 251ah       ; Install new int 1ah vector.
		lea     dx, newint1ah
		int     21h

		mov	ax, ds:[002ch]	; Attempt to free environment.
		cmp	ax, 0		; Can we free environment?
		jz	install6	; No, branch.

		mov	es, ax		; Free environment block.
		mov	ah, 49h
		int	21h

		mov	WORD PTR ds:[002ch], 0 ; Indicate no environment.
install6:
		mov     ah, 9           ; Display ok install message.
		lea     dx, okmsg1
		int     21h

		mov     ax, 3100h       ; TSR.
		lea     dx, res_end
		add     dx, 15
		mov     cl, 4
		shr     dx, cl
		int     21h

; ------------------------------------------------ ;
; Search: Search memory for previous TSR copy.     ;
;                                                  ;
; Entry: BX = address of signature to search       ;
;        CX = length of signature                  ;
;                                                  ;
; Exit:  AX = PSP address of prior or current copy ;
;                                                  ;
; All registers preserved.                         ;
; ------------------------------------------------ ;

search          PROC    NEAR
		push    bx
		push    cx
		push    si
		push    di
		push    es

		mov     si, bx          ; SI will point to signature.

		mov     ax, 0a000h      ; Initial segment.
		mov     bx, cs          ; Final segment.
search1:
		inc     ax              ; Point to next segment.

		cmp     ax, bx          ; At current segment?
		jz      search2         ; Yes, exit search.

		mov     es, ax          ; Target segment.
		mov     di, si          ; Same offset.

		push    cx              ; CMPSB destroys CX and SI.
		push    si

		rep     cmpsb           ; Search for match.

		pop     si
		pop     cx

		jnz     search1         ; Branch if no match.
search2:
		pop     es
		pop     di
		pop     si
		pop     cx
		pop     bx

		ret
search          ENDP

; ---------------------------------------- ;
; Uninstall: Attempt to uninstall SND TSR. ;
;                                          ;
; Exit: Carry-flag set if error.           ;
;                                          ;
; All registers preserved.                 ;
; ---------------------------------------- ;

uninstall       PROC    NEAR
		push    ax
		push    bx
		push    dx
		push    es

		mov     ax, 351ah       ; Get int 1ah vector.
		int     21h

		mov     ax, es          ; Get current vector segment.
		cmp     ax, psp         ; If not = to resident segment
		jnz     uninstall1      ; ... then another TSR has chained.

		mov     ax, 251ah       ; Restore old int 1ah vector.
		lds     dx, es:oldint1ah
		int     21h

		push    cs
		pop     ds

		mov     ah, 49h         ; Attempt to free memory block.
		int     21h

		jc      uninstall1      ; Branch if unsuccessful.

		clc                     ; Signal success.
		jmp     SHORT uninstall2
uninstall1:
		stc                     ; Signal failure.
uninstall2:
		pop     es
		pop     dx
		pop     bx
		pop     ax
		ret
uninstall       ENDP

_TEXT           ENDS
		END     snd

COMMENT ~

     Ŀ
                       PC TIMING AND SOUND GENERATION                  
                                                                       
      The heart of the PC timing mechanism is a crystal oscillator.    
      Circuitry is used to condition and divide the oscillator output  
      until it reaches the manufacturer's rated CPU speed (4.77 MHZ on 
      the original IBM PC).                                            
                                                                       
      The output frequency is further divided until it reaches 1.19318 
      MHZ.  This frequency is available to programmers.                
                                                                       
      The 1.19318 MHZ frequency is input to an Intel 8235 timer chip   
      (8254 on an AT or an ASIC on some compatibles).  This timer chip 
      has three different channels and each channel receives the same  
      input frequency.  These channels are labeled A, B, and C.  Each  
      channel has a gate that can allow the input frequency into the   
      channel or keep it out of the channel.  A pair of data latches,  
      a counter, and an output line finish off the channel.  All three 
      channels are controlled by a single control register.  The gates 
      for channels A and B are permanently open to allow the frequency 
      into these channels.  The channel C gate can be programmed to be 
      open or shut.  The following is a detailed schematic of a chan-  
      nel on the 8253/8254 timer chip.                                 
                                                                       
                        Ŀ Ŀ                   
                    Ĵ Clock input   Computer                    
                        1.19318 MHZ   I/O  BUS                    
                        ^                   
                                                     
                      ͻ                  
                        vĿ  Ŀ                    
                         Input      Output                      
                         Data       Data                        
                         Latch      Latch                       
                          ^                    
                    ĿĿ  Ŀ                      
                         vĿ  vĿ                    
                   > Gate    > Counter Ŀ              
                                             
                        8253 - One of 3 Channels                   
                      ͼ   v              
                Control                               Output           
                                                                       
      In most of its operating modes, a channel loads a value from its 
      input data latch into the counter, counts down toward 0 on each  
      input pulse, and outputs a pulse when the counter reaches 0.  It 
      then reloads the value into the counter and repeats as long as   
      the input gate is open.  This allows the input frequency to be   
      divided into a more convenient value.                            
                                                                       
      On the original IBM PC, channel A controls the real-time clock,  
      channel B controls the DRAM refresh circuitry, and channel C the 
      speaker.  Channel A is also used by a variety of disk drive con- 
      trollers.                                                        
                                                                       
      Each channel may operate in any one of six different modes (0-5) 
      which are summarized shortly.  Modes 2 and 3 are normally used.  
                                                                       
      The timer control register is located at port 43h.  This port is 
      write-only.  The read/write data latches are visible at ports    
      40h, 41h, and 42h for channels A, B, and C respectively.         
                                                                       
      To use the timer chip, the following sequence of operations must 
      be performed.                                                    
                                                                       
      1. Issue a command through the control port.                     
      2. Read or write values through the data latches.                
                                                                       
      The command to be executed is defined by bits 5 and 4 of the     
      control register.  The last three commands allow the counter to  
      be read directly without first having its value latched via the  
      output data latch register.  This approach causes a problem for  
      the last command which allows both bytes of the 16-bit counter   
      to be read: least significant byte followed by most significant. 
      Because the ports (41h-42h) are only 8-bits wide, two successive 
      reads are required.  After the least significant byte has been   
      read, it is possible for the counter to decrement by one or more 
      pulses before the most significant byte is read.  As a result,   
      the true counter value is distorted since the most significant   
      byte may be decremented during the count.  For example, suppose  
      the counter value is 400 (decimal) when the least significant    
      byte is read and then decrements to 399 before the most signif-  
      ificant byte is read.  The resulting value is 300, not the cor-  
      rect 400 or even 399.  This is the reason for the first command  
      which copies the counter to the output data latch.  Both reads   
      will return the latch value rather than the counter value which  
      results in accuracy.                                             
                                                                       
         Control Register: Port 43h                                    
             Ŀ                                         
             76543210                                         
                                                      
                                                                       
      Bits 7 and 6 select the counter to be used.  Bits 5 and 4 select 
      the read/load command.  Bits 3 and 2 and 1 select the counting   
      mode.  Bit 0 selects the type of counting (binary or BCD).       
                                                                       
      7 6                                                              
      - -                                                              
      0 0 - select channel A counter (port 40h)                        
      0 1 - select channel B counter (port 41h)                        
      1 0 - select channel C counter (port 42h)                        
      1 1 - illegal value                                              
                                                                       
      5 4                                                              
      - -                                                              
      0 0 - copy counter register to output data latch                 
      0 1 - read/load most significant byte only                       
      1 0 - read/load least significant byte only                      
      1 1 - read/load least significant byte followed by most signifi- 
            cant byte                                                  
                                                                       
      3 2 1 (all modes count from loaded value down to zero)           
      - - -                                                            
      0 0 0 - generate an interrupt when counter reaches 0             
      0 0 1 - programmable one-shot                                    
      0 1 0 - rate generator                                           
      0 1 1 - square wave generator: used by channel A and C (speaker) 
      1 0 0 - software triggered strobe                                
      1 0 1 - hardware triggered strobe                                
                                                                       
      Modes 2 and 3 are normally used on the PC.  Mode 2 counts down   
      to zero and triggers a 1-clock low pulse.  Mode 3 counts down by 
      2's setting the output low for an entire cycle and high for the  
      next cycle resulting in a square wave.                           
                                                                       
      If bit 0 is 0 then a 16-bit binary counter is used otherwise a   
      four-decade binary coded decimal (BCD) counter is used.          
                                                                       
      Channel B has its input gate wired open so it is always active.  
      This channel is responsible for maintaining DRAM refresh.  It is 
      of no interest in this discussion.                               
                                                                       
      During power-up, the BIOS loads the channel A input latch with a 
      value of 0 which is interpreted as 65536.  The input frequency   
      of 1.19318 MHZ divided by 65536 results in the 18.20648193359 HZ 
      signal that is associated with interrupt 8h (commonly known as   
      the timer tick interrupt).  About 18.2 times each second, the    
      channel A counter generates an IRQ0 hardware interrupt.  This    
      interrupt is fielded by the 8259 priority interrupt controller   
      resulting in interrupt 8h being activated.  Interrupt 8h has the 
      responsibility for updating the timer tick counter found at BIOS 
      addresses 0000:046Ch to 0000:046Fh.  Check out procedure delay   
      to see how we use this value.  The timer tick frequency can be   
      changed if used with care by loading a different value into the  
      channel A input latch.                                           
                                                                       
      There are two ways to generate sound from the speaker.  Toggling 
      bit 1 of port B under program control is one way.  When bit 1 is 
      set to 1 then current flows through the speaker.  When bit 1 is  
      set to 0 then current is blocked.  A program could set bit 1 and 
      sit in an idle loop for a certain number of cycles.  Then bit 1  
      would be set to 0.  The program would sit in another idle loop.  
      The length of these loops would determine the frequency of the   
      tone produced by the speaker.  Unfortunately, the tone frequency 
      is dependent upon the loop counters.  A loop running on an XT is 
      slower than a loop running on a 386 because of the higher clock  
      speed on the 386.  Therefore, a pleasant-sounding tone on the XT 
      would sound like a high-pitched whistle on a 386.  Also, these   
      loops would be interrupted approximately 18.2 times per second   
      by interrupt 8h resulting in distortion.  A third disadvantage   
      is that the computer is forced to spend all of its time handling 
      these loops.                                                     
                                                                       
      The second and better way of handling sound is:                  
                                                                       
      1. Turn off bits 0 and 1 of port B.  This blocks current to the  
         speaker and prevents channel C from sending pulses to port C. 
      2. Load the channel C input data latch with the frequency value. 
      3. Turn on bit 5 of port C (it's on by default) if necessary.    
         This allows channel C pulses to reach a tristate buffer just  
         in front of the speaker and controlled by bit 1 of port B.    
      4. Turn on bits 0 and 1 of port B.                               
      5. Go and do something else until the tone has lasted long       
         enough.                                                       
                                                                       
      The correct frequency value can be determined by dividing the    
      1.19318 MHZ frequency by the tone frequency desired.  An example 
      for a 1000 HZ tone is: 1,193,180/1000 = 1,193. Place 1,193 in    
      the channel C input data latch for this tone.                    
                                                                       
      Turning off the speaker is as simple as turning off bits 0 and 1 
      and the following schematic illustrates the speaker connection.  
                                                                       
                                    Ŀ Ŀ       
                                Ĵ Clock input   Computer        
                                    1.19318 MHZ   I/O  BUS        
                                    ^       
                                                     
                                  ͻ      
                                    vĿ  Ŀ        
                                     Input      Output          
                                     Data       Data            
                                     Latch      Latch           
                                      ^        
                                ĿĿ  Ŀ          
                                     vĿ  vĿ        
                              > Gate    > Counter Ŀ  
                                             
                                     Channel C - 8253 Timer        
                                  ͼ     
      ͻ                                   
                           Ŀ                          
       Port B                                     
        61h   76543210                                    
                   vĿ                     
                   Ŀ    SPEAKER                      
       Port C Ŀ    ^                     
        62h   76543210                           
              ^                                     
                     
       8255 Peripheral Interface                                     
      ͼ                                    
                                                                       
      Summary:                                                         
                                                                       
      1. Port B (61h) bit 0 controls channel C.  If set then the chan- 
         nel C counter will send its pulses to port C (62h) bit 5.     
         On some PCs, bit 5 of port C is permanently set to 1 so that  
         whenever channel C is active, its pulses are going to the     
         speaker as long as port B bit 1 is set.                       
                                                                       
      2. Port B (61h) bit 1 controls access to the speaker.  If set    
         then the speaker will receive current and can generate sound. 
         This sound is constant and the frequency is determined by     
         software timing loops.                                        
                                                                       
      3. Port C (62h) bit 5 controls access to the speaker by channel  
         C.  If set then the channel C pulses will reach the speaker.  
     

     Ŀ
      I am looking for technical data on the 1793 floppy disk control- 
      ler chip manufactured by Western Digital and found in the Radio  
      Shack color computer disk drive.  I need access to the instruc-  
      tions for programming this chip (e.g., reading sectors, format-  
      ting tracks, etc).  If anyone has any data on this chip, I would 
      appreciate a photocopy (Western Digital has no technical data).  
      My address is:                                                   
                                                                       
      Geoff Friesen                                                    
      General Delivery                                                 
      Dauphin, MB                                                      
      Canada R7N 2T3                                                   
                                                                       
      Thankyou for your help.                                          
     

~